#include "main.h"
#include <wininet.h>

extern HANDLE hNetlibUser;		// Handle of the Netlib User
extern char *pszModulename;		// Name of the module using Netlib
extern BOOL UseWinInet;			// Variable that indicates wether to use WinInet or Netlib

// Imports
	typedef BOOL  ( WINAPI *ft_HttpQueryInfo )( HINTERNET, DWORD, LPVOID, LPDWORD, LPDWORD );
	typedef BOOL  ( WINAPI *ft_HttpSendRequest )( HINTERNET, LPCSTR, DWORD, LPVOID, DWORD );
	typedef BOOL  ( WINAPI *ft_InternetCloseHandle )( HINTERNET );
	typedef DWORD ( WINAPI *ft_InternetErrorDlg )( HWND, HINTERNET, DWORD, DWORD, LPVOID* );
	typedef BOOL  ( WINAPI *ft_InternetSetOption )( HINTERNET, DWORD, LPVOID, DWORD );
	typedef BOOL  ( WINAPI *ft_InternetQueryOption)( HINTERNET, DWORD, LPVOID, LPDWORD );
	typedef BOOL  ( WINAPI *ft_InternetQueryDataAvailable)( HINTERNET, LPDWORD, DWORD, DWORD );

	typedef HINTERNET ( WINAPI *ft_HttpOpenRequest )( HINTERNET, LPCSTR, LPCSTR, LPCSTR, LPCSTR, LPCSTR*, DWORD, DWORD );
	typedef HINTERNET ( WINAPI *ft_InternetConnect )( HINTERNET, LPCSTR, INTERNET_PORT, LPCSTR, LPCSTR, DWORD, DWORD, DWORD );
	typedef HINTERNET ( WINAPI *ft_InternetOpen )( LPCSTR, DWORD, LPCSTR, LPCSTR, DWORD );
	typedef HINTERNET ( WINAPI *ft_InternetReadFile ) ( HINTERNET, LPVOID, DWORD, LPDWORD );

	static ft_InternetCloseHandle f_InternetCloseHandle;
	static ft_InternetConnect     f_InternetConnect;
	static ft_InternetErrorDlg    f_InternetErrorDlg;
	static ft_InternetOpen        f_InternetOpen;
	static ft_InternetSetOption   f_InternetSetOption;
	static ft_InternetQueryOption f_InternetQueryOption;
	static ft_InternetReadFile	  f_InternetReadFile;
	static ft_HttpOpenRequest     f_HttpOpenRequest;
	static ft_HttpQueryInfo       f_HttpQueryInfo;
	static ft_HttpSendRequest     f_HttpSendRequest;
	static ft_InternetQueryDataAvailable     f_InternetQueryDataAvailable;


//WinInet Functions

#ifdef USE_WININET
int LoadWinInet(void) {
	HMODULE tDll;

	if (!(tDll=LoadLibrary("WinInet.dll"))) return -1;
	if (!(f_InternetCloseHandle = (ft_InternetCloseHandle)GetProcAddress( tDll, "InternetCloseHandle" ))||
	    !(f_InternetConnect = (ft_InternetConnect)GetProcAddress( tDll, "InternetConnectA" ))			||
	    !(f_InternetErrorDlg = (ft_InternetErrorDlg)GetProcAddress( tDll, "InternetErrorDlg" ))			||
	    !(f_InternetOpen = (ft_InternetOpen)GetProcAddress( tDll, "InternetOpenA" ))					||
	    !(f_InternetSetOption = (ft_InternetSetOption)GetProcAddress( tDll, "InternetSetOptionA" ))		||
		!(f_InternetQueryOption = (ft_InternetQueryOption)GetProcAddress( tDll, "InternetQueryOptionA"))||
	    !(f_HttpOpenRequest = (ft_HttpOpenRequest)GetProcAddress( tDll, "HttpOpenRequestA" ))			||
	    !(f_HttpQueryInfo = (ft_HttpQueryInfo)GetProcAddress( tDll, "HttpQueryInfoA" ))					||
	    !(f_HttpSendRequest = (ft_HttpSendRequest)GetProcAddress( tDll, "HttpSendRequestA" ))			||
		!(f_InternetReadFile = (ft_InternetReadFile)GetProcAddress( tDll, "InternetReadFile" ))			||
		!(f_InternetQueryDataAvailable = (ft_InternetQueryDataAvailable)GetProcAddress(tDll, "InternetQueryDataAvailable"))) {
		  OUTPUT("Cannot import all needed functions from WinInet, NETLIB will be used instead.\n");
		  if (( GetVersion() & 0x80000000 ) == 0) FreeLibrary(tDll);  // Unload WinInet under NT 4.0
		  return -1;
	}
	return 0;
}


/* Simulates a NETLIBHTTPTRANSACTION by using the WinInet functions
   The following things will be currently ignored, as they are already
   generated by the WinInet API:

   * Header fields (except referer, Cookies which are extracted)

  This function was introduced in V0.0.0.5 to be able to to SSL transactions,
  however plain NETLIB-requests are still supported.
  You may compile this program without WinInet support at any time so
  that this bloated, IE-depending lib is not needed, however, SSL-
  transactions will not be supported then.
 */
int WinInetGet(NETLIBHTTPREQUEST *nlhr)
{
	DWORD tFlags =
		INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTPS |
		INTERNET_FLAG_IGNORE_REDIRECT_TO_HTTP |
		INTERNET_FLAG_IGNORE_CERT_CN_INVALID |
		INTERNET_FLAG_IGNORE_CERT_DATE_INVALID |
		INTERNET_FLAG_NO_CACHE_WRITE |
		INTERNET_FLAG_RELOAD |
		INTERNET_FLAG_NO_UI |
		INTERNET_FLAG_PRAGMA_NOCACHE |
		INTERNET_FLAG_NO_COOKIES |
		INTERNET_FLAG_NO_AUTO_REDIRECT;

//		//INTERNET_FLAG_NO_COOKIES  --> Let WININET manage our Cookies, but we still
//									  manage them ourselves for the $Cookie[] vars.

	HINTERNET tNetHandle, tUrlHandle;
	BYTE tUsesProxy;
	DBVARIANT dbv;
	DWORD AccessType=INTERNET_OPEN_TYPE_PRECONFIG, tErrorCode;
	INTERNET_PORT dwPort=INTERNET_DEFAULT_HTTP_PORT;
	NETLIBHTTPREQUEST *nlhrResponse=NULL;
	char tBuffer[SSL_BUF_SIZE], *urlStart, *tObjectName, *host=NULL, *reqType=NULL, *Referer=NULL, *p=NULL, *UserAgent=pszModulename, *header=NULL;
	int i;

	// Use Proxy from Netlib, if needed.
	// Comment this section out if you always want to use IE-default
	if ( (tUsesProxy=DBGetContactSettingByte(NULL, pszModulename, "NLUseProxy", 0)))	{
		int tPortNumber;

		memset(tBuffer, 0, SSL_BUF_SIZE);
		if (!DBGetContactSetting( NULL, pszModulename, "NLProxyServer", &dbv ))
		{	
			if ((tPortNumber=DBGetContactSettingWord(NULL, pszModulename, "NLProxyPort", -1 ))>0) {
			  AccessType=INTERNET_OPEN_TYPE_PROXY;
			  _snprintf( tBuffer, SSL_BUF_SIZE, "%s:%d", dbv.pszVal, tPortNumber);
			}
			DBFreeVariant(&dbv);
		}
	}
	else AccessType=INTERNET_OPEN_TYPE_DIRECT;

	// Set Request type
	switch (nlhr->requestType) {
		case REQUEST_GET: reqType="GET"; break;
		case REQUEST_POST: reqType="POST"; header=strdup("Content-Type: application/x-www-form-urlencoded"); break;
		case REQUEST_CONNECT: reqType="CONNECT"; break;
	}

	// Extract Referer, User-Agent, Cookies, ...
	for (i=0; i<nlhr->headersCount; i++) {
		if (!stricmp(nlhr->headers[i].szName, "Referer"))
			Referer=nlhr->headers[i].szValue;
		if (!stricmp(nlhr->headers[i].szName , "User-Agent"))
			UserAgent=nlhr->headers[i].szValue;
		if (!stricmp(nlhr->headers[i].szName, "Cookie")) {
			if (header) {
				header=realloc(header, strlen(header)+strlen(nlhr->headers[i].szName)+strlen(nlhr->headers[i].szValue)+3+strlen("\n"));
				strcat(header, "\n");
			} else {
				header=malloc(strlen(nlhr->headers[i].szName)+strlen(nlhr->headers[i].szValue)+3);
				header[0]=0;
			}
			strcat(header, nlhr->headers[i].szName);
			strcat(header, ": ");
			strcat(header, nlhr->headers[i].szValue);
		}
	}
	DEBUG_OUT("Additional Headers: %s\n", header);

	if (!(tNetHandle=f_InternetOpen(UserAgent, AccessType, tBuffer, NULL, 0)))
		return 0;

	DEBUG_OUT( "HTTP request (%s): '%s'\n", ( tUsesProxy ) ? "using proxy": "direct connection", nlhr->szUrl );


	// Split URL in Host and Object portions and decide if HTTPS or HTTP has to be used
	if (!(urlStart = strstr( nlhr->szUrl, "://" ))) urlStart = nlhr->szUrl;
	else {
		if (stristr(nlhr->szUrl, "https")) {
			dwPort=INTERNET_DEFAULT_HTTPS_PORT;
			tFlags|=INTERNET_FLAG_SECURE;
		}
		urlStart += 3;
	}

	if (tObjectName = ( char* )strchr(urlStart, '/')) {
		tObjectName[0]=0;
		host=strdup(urlStart);
		// Seperate Port from host-portion if it exists
		if (p=strchr(host, ':')) {
			p[0]=0;
			dwPort=atoi(p+1);
		}
		tObjectName[0]='/';
	} else tObjectName = "/";
	
	
	// Let's rock
	if (tUrlHandle = f_InternetConnect(tNetHandle, host, dwPort, "", "", INTERNET_SERVICE_HTTP, tFlags, 0 ) ) {
		HINTERNET tRequest = f_HttpOpenRequest(tUrlHandle, reqType, tObjectName, NULL, Referer, NULL, tFlags, 0 );
		if (tRequest) {
			DWORD dwFlags;
			DWORD tBufSize;

			if (tUsesProxy) {
				OUTPUT( "Applying proxy parameters...\n");

				if (!DBGetContactSetting( NULL, pszModulename, "NLProxyAuthUser", &dbv )) {
					f_InternetSetOption( tRequest, INTERNET_OPTION_PROXY_USERNAME, dbv.pszVal, strlen(dbv.pszVal)+1);
					DBFreeVariant(&dbv);
				}

				if (!DBGetContactSetting( NULL, pszModulename, "NLProxyAuthPassword", &dbv )) {
					f_InternetSetOption( tRequest, INTERNET_OPTION_PROXY_PASSWORD, dbv.pszVal, strlen(dbv.pszVal)+1);
					DBFreeVariant(&dbv);
				}
			}
			// Ignore certificate stuff for HTTPS-Transactions
			tBufSize=sizeof(dwFlags);
			f_InternetQueryOption (tRequest, INTERNET_OPTION_SECURITY_FLAGS, (LPVOID)&dwFlags, &tBufSize);
			dwFlags |= SECURITY_FLAG_IGNORE_UNKNOWN_CA;
			f_InternetSetOption (tRequest, INTERNET_OPTION_SECURITY_FLAGS, &dwFlags, sizeof (dwFlags) );

LBL_Restart:
			if ( !(tErrorCode=f_HttpSendRequest(tRequest, header, header?strlen(header):0, nlhr->pData, nlhr->dataLength)) ) {
				tErrorCode = GetLastError();
				DEBUG_OUT( "HttpSendRequest() failed with error %ld", tErrorCode );

				if ( tErrorCode == 2 )
					SHOWERROR("Internet Explorer is in the 'Offline' mode. Switch IE to the 'Online' mode and then try to relogin");
			} else {
				DWORD dwCode;
				tBufSize = sizeof( dwCode );
				f_HttpQueryInfo( tRequest, HTTP_QUERY_STATUS_CODE | HTTP_QUERY_FLAG_NUMBER, &dwCode, &tBufSize, 0 );

				switch( dwCode ) {
				case ERROR_INTERNET_HTTP_TO_HTTPS_ON_REDIR:
				case ERROR_INTERNET_INCORRECT_PASSWORD:
				case ERROR_INTERNET_INVALID_CA:
				case ERROR_INTERNET_POST_IS_NON_SECURE:
				case ERROR_INTERNET_SEC_CERT_CN_INVALID:
				case ERROR_INTERNET_SEC_CERT_DATE_INVALID:
					DEBUG_OUT("There was an error in the resultcode for the request: %d\n", dwCode);
					if ( ERROR_CANCELLED != f_InternetErrorDlg( GetDesktopWindow(), tRequest, tErrorCode, FLAGS_ERROR_UI_FILTER_FOR_ERRORS | FLAGS_ERROR_UI_FLAGS_GENERATE_DATA | FLAGS_ERROR_UI_FLAGS_CHANGE_OPTIONS, NULL ))
						goto LBL_Restart;

					// else fall into the general error handling routine
					break;
				default:

					nlhrResponse=(NETLIBHTTPREQUEST*)calloc(1,sizeof(NETLIBHTTPREQUEST));
					nlhrResponse->cbSize=sizeof(NETLIBHTTPREQUEST);
					nlhrResponse->requestType=REQUEST_RESPONSE;
					nlhrResponse->resultCode=dwCode;
					
					tBufSize = SSL_BUF_SIZE;
					if ( !f_HttpQueryInfo( tRequest, HTTP_QUERY_STATUS_TEXT, tBuffer, &tBufSize, NULL ))
						strcpy( tBuffer, "unknown error" );
					nlhrResponse->szResultDescr=strdup(tBuffer);
					DEBUG_OUT("HTTP query status code: %d - %s\n", dwCode, tBuffer);

					tBufSize = sizeof(nlhrResponse->dataLength);
					f_HttpQueryInfo( tRequest, HTTP_QUERY_CONTENT_LENGTH | HTTP_QUERY_FLAG_NUMBER, &nlhrResponse->dataLength, &tBufSize, 0 );
					DEBUG_OUT("Content length: %d\n", nlhrResponse->dataLength);

					if (!nlhrResponse->dataLength) {
						char *endptr;
						DWORD nRemaining;
						// Server didn't tell us Content-Length so we have to downlaod it
						// chunk-by-cunk.. This basically SUX!! :(
						while (f_InternetQueryDataAvailable(tRequest, &nRemaining, 0, 0 ) && nRemaining > 0 ) {	
							DEBUG_OUT("Data available, %d bytes available\n", nRemaining);
							if (!(nlhrResponse->pData=(char *)realloc(nlhrResponse->pData, nlhrResponse->dataLength+nRemaining+1))) {
								nlhrResponse->dataLength=0;
								OUTPUT("ERROR: Out of memory while allocating space for Data\n");
								break;
							}
							endptr=nlhrResponse->pData+nlhrResponse->dataLength;
							ZeroMemory(endptr, nRemaining+1);
							if (!f_InternetReadFile(tRequest, endptr, nRemaining, &nRemaining))
							{
								OUTPUT("ERROR: InternetReadFile failed with error 0x%.8X\n", GetLastError());
								nlhrResponse->dataLength=0;
								break;
							}
							nlhrResponse->dataLength+=nRemaining;
						}
						//if (!nlhrResponse->dataLength) break;
					} else {
						nlhrResponse->pData=(char *)calloc(1, nlhrResponse->dataLength+1);
						if (!f_InternetReadFile( tRequest, nlhrResponse->pData, nlhrResponse->dataLength, &tBufSize))
						{
							OUTPUT("ERROR: InternetReadFile failed with error 0x%.8X\n", GetLastError());
							break;
						}
					}
					DEBUG_OUT("InternetReadFile read %d bytes\n", tBufSize);
					tBufSize = SSL_BUF_SIZE-1;
					ZeroMemory(tBuffer, tBufSize+1);
					f_HttpQueryInfo( tRequest, HTTP_QUERY_RAW_HEADERS, tBuffer, &tBufSize, NULL );
					DEBUG_OUT("Length of raw headers: %d\n", tBufSize);
					p=tBuffer;
					while (p[0]) {
						char *p2;
						unsigned int plen;

						nlhrResponse->headers=(NETLIBHTTPHEADER*)realloc(nlhrResponse->headers,sizeof(NETLIBHTTPHEADER)*(nlhrResponse->headersCount+1));
						plen=strlen(p);
						if (p2=strchr(p, ':')) {
							p2[0]=0;
							nlhrResponse->headers[nlhrResponse->headersCount].szName=strdup(p);
							while(*(++p2)==' ' || *p2=='\t');
							nlhrResponse->headers[nlhrResponse->headersCount++].szValue=strdup(p2);
							p2[0]=':';
						}
						p+=plen+1;
					}
					DEBUG_OUT("Raw headers read.\n");
					
					break;
				}
			}
			if (header) free(header);
			f_InternetCloseHandle( tRequest );
		}

		f_InternetCloseHandle( tUrlHandle );
	}
	else {DEBUG_OUT ( "InternetOpenUrl() failed\n" );}

	f_InternetCloseHandle( tNetHandle );
	if (host) free(host);
	return (int)nlhrResponse;
}
#endif

int _HttpFreeRequestStruct(NETLIBHTTPREQUEST *nlhr)
{
#ifdef USE_WININET
	if (UseWinInet) {
		// Free nlhr on local heap
		if(nlhr==NULL || nlhr->cbSize!=sizeof(NETLIBHTTPREQUEST) || nlhr->requestType!=REQUEST_RESPONSE) {
			SetLastError(ERROR_INVALID_PARAMETER);
			return 0;
		}
		if(nlhr->headers) {
			int i;
			for(i=0;i<nlhr->headersCount;i++) {
				if(nlhr->headers[i].szName) free(nlhr->headers[i].szName);
				if(nlhr->headers[i].szValue) free(nlhr->headers[i].szValue);
			}
			free(nlhr->headers);
		}
		if(nlhr->pData) free(nlhr->pData);
		if(nlhr->szResultDescr) free(nlhr->szResultDescr);
		if(nlhr->szUrl) free(nlhr->szUrl);
		free(nlhr);
		return 1;	
	} else
#endif
	return CallService(MS_NETLIB_FREEHTTPREQUESTSTRUCT,0,(LPARAM)nlhr);
}


NETLIBHTTPREQUEST* _HTTPtransaction(NETLIBHTTPREQUEST *nlhr) {
	DEBUG_OUT("Calling MS_NETLIB_HTTPTRANSACTION with handle %d\n", hNetlibUser);
#ifdef USE_WININET
			if (UseWinInet) return (NETLIBHTTPREQUEST*)WinInetGet(nlhr); else
#endif
			return (NETLIBHTTPREQUEST*)CallService(MS_NETLIB_HTTPTRANSACTION,(WPARAM)hNetlibUser,(LPARAM)nlhr);
	DEBUG_OUT("Transaction complete.\n");
}
